/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Duesterloh <mnenhy@tprac.de>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2002-2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Karsten Duesterloh <mnenhy@tprac.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// +-
// |    Written, forged, copied and conceived by Karsten Duesterloh <mnenhy@tprac.de>
// |    This file is inspired by
// |      MozTweakRDF.js by Marek Wawoczny <Marek.Wawoczny@GmbH.rubikon.pl>, and
// |      pRDFData.js by SHIMODA Hiroshi <piro@p.club.ne.jp>
// |    But there isn't too much code left anyway that has survived my editing... ;-)
// +-

// +-
// |  the contents of this file will be loaded into the Mnenhy.rdf.* context!
// |  ie. foo() is available as goMnenhy.rdf.foo()
// +-

// |  The rough structure of an RDF file as seen by this script:
// |
// |   <rdf:RDF>
// |     <rdf:$container about="$urn">
// |       <rdf:$node about="$urn">
// |         <$datanamespace:$item>$value</$datanamespace:$item>
// |         ...
// |       </rdf:$node>
// |       ...
// |     </rdf:$container>
// |     ...
// |   </rdf:RDF>
// |
// |  $container = "Alt" | "Bag" | "Seq"
// |  $node      = $container | "Description"
// |  $urn is a URN like "urn:mozilla:overlays" and should be unique
// |  $datanamespace is a namespace like "http://www.mozilla.org/rdf/chrome#"
// |  $item is an element in $datanamespace whose $value is stored in the rdf


// |
// |  constants
// |

// |  load state
const gkiBitMnenhyRDF_Clear  = 0x00;
const gkiBitMnenhyRDF_Load   = 0x01;
const gkiBitMnenhyRDF_Halt   = 0x02;
const gkiBitMnenhyRDF_Ready  = 0x04;
const gkiBitMnenhyRDF_Error  = 0x08;
// |  container types (MUST NOT be 0)
const gkiRDFContAlt          = 1;
const gkiRDFContBag          = 2;
const gkiRDFContSeq          = 3;
// |  interfaces
const gkIRDFRemoteDataSource = Components.interfaces.nsIRDFRemoteDataSource;
const gkIRDFXMLSink          = Components.interfaces.nsIRDFXMLSink;
const gkIRDFService          = Components.interfaces.nsIRDFService;
const gkIRDFContainer        = Components.interfaces.nsIRDFContainer;
const gkIRDFContainerUtils   = Components.interfaces.nsIRDFContainerUtils;
const gkIRDFNode             = Components.interfaces.nsIRDFNode;
const gkIRDFResource         = Components.interfaces.nsIRDFResource;
const gkIRDFLiteral          = Components.interfaces.nsIRDFLiteral;
//const gkIRDFDate             = Components.interfaces.nsIRDFDate;
//const gkIRDFInt              = Components.interfaces.nsIRDFInt;
//const gkIRDFBlob             = Components.interfaces.nsIRDFBlob;


// |
// |  class MnenhyRDF
// |

// Constructor
function MnenhyRDF(
/* string or nsIURI */ aoDataSourceFile,              // URI of rdf-file
    /* object   */ aoNotifier          /* = DefaultNotifier */, // (see below for explanation)
    /* string   */ asDataNameSpace     /* = ""    */, // namespace for elements inside of rdf:description
/* string-array */ arsURNPrefix        /* = []    */, // prefix to be added to all URNs, eg. [urn,mozilla]
    /* boolean  */ abBlocking          /* = false */, // if true, rdf is opened synchronously
    /* nsIURI   */ aoDefaultSourceFile /* = null  */, // URI of default rdf file
    /* boolean  */ abEnableDebugDump   /* = false */  // debugdump einschalten
)
{
  this.moRDFService            = Components.classes["@mozilla.org/rdf/rdf-service;1"]    .getService(gkIRDFService);
  this.moRDFContainerUtils     = Components.classes["@mozilla.org/rdf/container-utils;1"].getService(gkIRDFContainerUtils);
  this.mCRDFContainer          = new Components.Constructor("@mozilla.org/rdf/container;1", "nsIRDFContainer", "Init");

  this.mbDebugDump             = false;

  this.mbBlocking              = abBlocking || false;
  this.moDataSource            = null;
  this.mbDataSourceLoaded      = false;
  this.mbDataSourceCreated     = false;   // only true, if new rdf file has been created
  this.miDataSourceState       = gkiBitMnenhyRDF_Clear;
  this.msDataSourceURI         = this.GetDataSourceURI(aoDataSourceFile, aoDefaultSourceFile);
  this.mSinkObserverRegistered = false;

  // If the datasource is ready for usage, a notifier function will be called:
  //    this.mfNotifier(this MnenhyRDF object, this.moUserData);
  // In case of error, the MnenhyRDF object will be |null|.
  // If aoNotifier is a function, this.moUserData is |null|,
  // otherwise it must be an array [fNotifier, oUserData]!
  this.moUserData = null;
  this.mfNotifier = DefaultNotifier;
  if (aoNotifier)
  {
    if (typeof(aoNotifier) == "function")
    { // no userdata
      this.mfNotifier = aoNotifier;
    }
    else if ("length" in aoNotifier)
    { // have array
      if (0 in aoNotifier)
        this.mfNotifier = aoNotifier[0];
      if (1 in aoNotifier)
        this.moUserData = aoNotifier[1];
    }
  }

  this.msDataNameSpace         = asDataNameSpace || "";
  this.mrsURNPrefix            = [];
  this.SetURNPrefix(arsURNPrefix || []);

  // enable debugdump
  if (abEnableDebugDump)
    this.EnableDebugDump();

  this.LoadDataSource();
}


// the prototype for the RDF object
MnenhyRDF.prototype =
{
  // |
  // |  datasource functions
  // |

  LoadDataSource: function()
  { // we only need to load the datasource if it hasn't been done already
    if(!this.moDataSource || !this.mbDataSourceLoaded || !this.mSinkObserverRegistered)
    {
      // synchronous or asynchronous?
      try
      {
        if (this.mbBlocking)
        {
          this.moDataSource = this.moRDFService.GetDataSourceBlocking(this.msDataSourceURI);
          // if the datasource is already cached but not loaded: refresh!
          // (we need this to sweep e.g. the old overlayinfo)
          if ((this.moDataSource instanceof gkIRDFRemoteDataSource) && (!this.moDataSource.loaded))
            this.RefreshDataSource();
        }
        else
          this.moDataSource = this.moRDFService.GetDataSource(this.msDataSourceURI);
      }
      catch(oError)
      { // usually we get here, if the URI does not exist
        dump("LoadDataSource:\n#"+this.msDataSourceURI+"#\n"+oError + "\n")
        this.miDataSourceState = gkiBitMnenhyRDF_Error;
        this.mfNotifier(null, this.moUserData);
        return;
      }

      var sinkObserver =
      {
        RDF:          this,
        onBeginLoad:  function(aSink)
                      {
                        this.RDF.debugdump("onBeginLoad:"+this.RDF.msDataSourceURI+"\n");
                        this.RDF.miDataSourceState = (this.RDF.miDataSourceState | gkiBitMnenhyRDF_Load);
                      },
        onInterrupt:  function(aSink)
                      {
                        this.RDF.debugdump("onInterrupt:"+this.RDF.msDataSourceURI+"\n");
                        this.RDF.miDataSourceState = (this.RDF.miDataSourceState | gkiBitMnenhyRDF_Halt);
                      },
        onResume:     function(aSink)
                      {
                        this.RDF.debugdump("onResume   :"+this.RDF.msDataSourceURI+"\n");
                        this.RDF.miDataSourceState = (this.RDF.miDataSourceState & ~gkiBitMnenhyRDF_Halt);
                      },
        onError:      function(aSink, aStatus, aErrorMsg)
                      {
                        this.RDF.debugdump("onError    :"+this.RDF.msDataSourceURI+"\n");
                        this.RDF.miDataSourceState = (this.RDF.miDataSourceState | gkiBitMnenhyRDF_Error);
                        //alert(aStatus + ": " + aErrorMsg);
                        this.RDF.mfNotifier(null, this.RDF.moUserData);
                        this.RDF.KillSink(this);
                      },
        onEndLoad:    function(aSink)
                      {
                        this.RDF.debugdump("onEndLoad  :"+this.RDF.msDataSourceURI+"\n");
                        this.RDF.miDataSourceState = (this.RDF.miDataSourceState | gkiBitMnenhyRDF_Ready);
                        this.RDF.mbDataSourceLoaded = (this.RDF.miDataSourceState == (gkiBitMnenhyRDF_Load + gkiBitMnenhyRDF_Ready));
                        this.RDF.mfNotifier(this.RDF, this.RDF.moUserData);
                        this.RDF.KillSink(this);
                      }
      }

      if ((this.moDataSource instanceof gkIRDFRemoteDataSource) && (this.moDataSource.loaded))
      {
        this.mbDataSourceLoaded = (this.moDataSource != null);
        this.mfNotifier(this, this.moUserData);
      }
      else
      { // register sink observer
        this.mbDataSourceLoaded       = false;
        this.miDataSourceState        = 0x0;
        if (this.moDataSource instanceof gkIRDFXMLSink)
        {
          this.moDataSource.addXMLSinkObserver(sinkObserver);
          this.mSinkObserverRegistered = true;
        }
        else
        {
          this.mSinkObserverRegistered = false;
        }
      } // register sink observer
    } // init
  },

  // Killing the sink observer because it would *leak* otherwise!
  KillSink:function(aoSink)
  {
    this.moDataSource.removeXMLSinkObserver(aoSink);
  },


  IsLoading: function()
  { // no error, no ready, but loading
    return ( (this.miDataSourceState & ~gkiBitMnenhyRDF_Halt) == gkiBitMnenhyRDF_Load);
  },
  IsGood: function()
  { // data source loaded successfully
    return this.mbDataSourceLoaded;
  },
  IsBad: function()
  { // error
    return ( (this.miDataSourceState & gkiBitMnenhyRDF_Error) != gkiBitMnenhyRDF_Clear);
  },


  RefreshDataSource: function(abBlocking)
  {
    const kbBlocking = (abBlocking !== null) ? abBlocking : this.mbBlocking;
    this.debugdump("RefreshDataSource("+abBlocking+"): blocking="+kbBlocking+" "+this.msDataSourceURI+"\n");
    if (this.moDataSource instanceof gkIRDFRemoteDataSource)
      this.moDataSource.Refresh(kbBlocking);
  },


  // write the current content to disk
  // if the string/nsIURI aoURI is given, write it there via FlushTo,
  // else just use Flush
  FlushDataSource: function(aoURI)
  {
    this.debugdump("FlushDataSource:");
    var sURI = aoURI;
    if (sURI instanceof Components.interfaces.nsIURI)
      sURI = sURI.spec;
    if (sURI)
    {
      this.debugdump(sURI+"\n");
      if (this.moDataSource instanceof gkIRDFRemoteDataSource)
      {
        this.moDataSource.FlushTo(sURI);
        this.debugdump("done\n");
      }
    }
    else
    {
      this.debugdump(this.msDataSourceURI+"\n");
      if (this.moDataSource instanceof gkIRDFRemoteDataSource)
      {
        this.moDataSource.Flush();
        this.debugdump("done\n");
      }
    }
  },


  // |
  // |  container functions
  // |

  // set the URN prefix
  // this prefix will be added to all subsequent node names
  SetURNPrefix: function(arsURNPrefix)
  {
    const koOldPrefix = this.mrsURNPrefix.slice();  // copy current value
    this.mrsURNPrefix = arsURNPrefix || [];         // set new value
    return koOldPrefix;
  },


  // create URN string
  // If aoParts is a string, it will be splitted into an array first;
  //  so that CreateURN("a:b") == CreateURN(["a","b"])
  // The aoParts array will be converted to an URN string,
  //  if abRelative is |true|, the URN base is skipped.
  // Empty parts will be ignored
  CreateURN: function(aoParts, abRelative)
  {
    if (!aoParts)
      aoParts = [];
    else if (typeof(aoParts) == "string")
      aoParts = aoParts.split(":");
    var r = aoParts.slice();  // clone aoParts
    if (!abRelative)
      r = this.mrsURNPrefix.concat(r);
    // replace multiple : by one :, delete leading and trailing :
    return r.join(":").replace(/:+/g,":").replace(/(^:+)|(:+$)/g,"");
  },


  // get the container element count
  GetContainerSize: function(arsContainer)
  { // we can't use nsIRDFContainer::GetCount, 
    // because that stupid function only returns the next RDF index minus one :-/
    var iSize = 0;
    const krsNodes = this.GetNodeList(arsContainer);
    if (krsNodes instanceof Array)
      iSize = krsNodes.length;
    return iSize;
  },


  // create a new container
  // if arsParentContainer is given, aoContainer will be a child node of it
  // if no arsParentContainer is given, the new container will be at top level of file
  // (the name of aoContainer is independent from arsParentContainer!)
  // aiContainerType must be of type gkiRDFCont*
  // aoContainer and aoParentContainer should be urn array or nsIRDFResource
  //
  // Note:
  // creating a new top level <rdf:description> can be done like this
  //    this.moDataSource.Assert(this.GetRes("about-name"),
  //                            this.GetRes("http://www.w3.org/1999/02/22-rdf-syntax-ns#instanceOf"),
  //                            this.GetRes("http://www.w3.org/1999/02/22-rdf-syntax-ns#Description"),
  //                            true);
  AddContainer: function(aoContainer, aiContainerType, aoParentContainer)
  { // create container resource
    if (!aoContainer)
      return;
    const koContainer = this.GetNodeRes(aoContainer);
    // check parent if given
    var oParent = null;
    if (aoParentContainer && !(oParent = this.InitContainer(aoParentContainer)))
      return; // invalid parent

    switch (aiContainerType)
    {
      case gkiRDFContAlt:
        this.moRDFContainerUtils.MakeAlt(this.moDataSource, koContainer);
        break;
      case gkiRDFContBag:
        this.moRDFContainerUtils.MakeBag(this.moDataSource, koContainer);
        break;
      case gkiRDFContSeq:
      default:
        this.moRDFContainerUtils.MakeSeq(this.moDataSource, koContainer);
        break;
    }
    // add new container to parent if necessary
    if (oParent && koContainer && (oParent.IndexOf(koContainer) == -1) )
      oParent.AppendElement(koContainer);
  },


  // get all arcs pointing into the given nsIRDFNode aoNode
  // and call afCallback(oSource, oArc, aoNode, aoUserData) for each one (if given)
  // if afCallback returns |false|, looping the arcs will be stopped!
  GetAllArcsIn: function(aoNode, afCallback, aoUserData)
  {
    this.debugdump(",- GetAllArcsIn\n");
    if (aoNode instanceof gkIRDFNode)
    {
      var bContinue = true;
      var oArcLabelsIn = this.moDataSource.ArcLabelsIn(aoNode);
      while(bContinue && oArcLabelsIn.hasMoreElements())
      {
        var oArc = oArcLabelsIn.getNext();
        var oSources = this.moDataSource.GetSources(oArc, aoNode, true);
        while (bContinue && oSources.hasMoreElements())
        {
          var oSource = oSources.getNext();

          // debug output!
          if (this.mbDebugDump)
          {
            const ks = ((oSource instanceof gkIRDFResource) ? "(R) "
                     : ((oSource instanceof gkIRDFLiteral)  ? "(L) "
                     : "    ")) + oSource.Value;
            const ka = ((oArc    instanceof gkIRDFResource) ? "(R) "
                     : ((oArc    instanceof gkIRDFLiteral)  ? "(L) "
                     : "    ")) + oArc.Value;
            const kt = ((aoNode  instanceof gkIRDFResource) ? "(R) "
                     : ((aoNode  instanceof gkIRDFLiteral)  ? "(L) "
                     : "    ")) + aoNode.Value;
            this.debugdump("| Source="+ks+"\n");
            this.debugdump("| Arc   ="+ka+"\n");
            this.debugdump("| Target="+kt+"\n");
            this.debugdump("+-\n");
          }

          if (afCallback)
            bContinue = afCallback(oSource, oArc, aoNode, aoUserData);

        } // oSource loop
      } // oArc loop
    } // aoNode is resource
    this.debugdump("`---------------\n");
  },


  // get all arcs pointing from the given nsIRDFResource aoResource
  // and call afCallback(aoResource, oArc, oTarget, aoUserData) for each one (if given)
  // if afCallback returns |false|, looping the arcs will be stopped!
  GetAllArcsOut: function(aoResource, afCallback, aoUserData)
  {
    this.debugdump(",- GetAllArcsOut\n");
    if (aoResource instanceof gkIRDFResource)
    {
      var bContinue = true;
      var oArcLabelsOut = this.moDataSource.ArcLabelsOut(aoResource);
      while(bContinue && oArcLabelsOut.hasMoreElements())
      {
        var oArc = oArcLabelsOut.getNext();
        var oTargets = this.moDataSource.GetTargets(aoResource, oArc, true);
        while (bContinue && oTargets.hasMoreElements())
        {
          var oTarget = oTargets.getNext();

          // debug output!
          if (this.mbDebugDump)
          {
            const ks = ((aoResource instanceof gkIRDFResource) ? "(R) "
                     : ((aoResource instanceof gkIRDFLiteral)  ? "(L) "
                     : "    ")) + aoResource.Value;
            const ka = ((oArc       instanceof gkIRDFResource) ? "(R) "
                     : ((oArc       instanceof gkIRDFLiteral)  ? "(L) "
                     : "    ")) + oArc.Value;
            const kt = ((oTarget    instanceof gkIRDFResource) ? "(R) "
                     : ((oTarget    instanceof gkIRDFLiteral)  ? "(L) "
                     : "    ")) + oTarget.Value;
            this.debugdump("| Source="+ks+"\n");
            this.debugdump("| Arc   ="+ka+"\n");
            this.debugdump("| Target="+kt+"\n");
            this.debugdump("+-\n");
          }

          if (afCallback)
            bContinue = afCallback(aoResource, oArc, oTarget, aoUserData);

        } // oTarget loop
      } // oArc loop
    } // aoResource is resource
    this.debugdump("`---------------\n");
  },


  // remove a single arc from the rdf datasource
  RemoveArc: function(aoResource, aoArc, aoTarget, aoDataSource)
  {
    if (    (aoResource instanceof gkIRDFResource)
         && (aoArc      instanceof gkIRDFResource)
         && (aoTarget   instanceof gkIRDFNode)
       )
    {
      aoDataSource.Unassert(aoResource, aoArc, aoTarget);
      return true;  // continue arc looping
    }
  },


  // remove all outgoing arcs from nsIRDFNode aoNode
  RemoveArcsIn: function(aoNode)
  {
    this.GetAllArcsIn(aoNode, this.RemoveArc, this.moDataSource);
  },


  // remove all outgoing arcs from nsIRDFResource aoResource
  RemoveArcsOut: function(aoResource)
  {
    this.GetAllArcsOut(aoResource, this.RemoveArc, this.moDataSource);
  },


  // remove container with all its elements
  // aoContainer should be urn array or nsIRDFResource
  // call afCallback(koElement, aoUserData) for each one (if given)
  RemoveContainer: function(aoContainer, afCallback, aoUserData)
  { // check existence
    const koContainer = this.GetNodeRes(aoContainer);
    var oCont;
    if (!(oCont = this.InitContainer(koContainer)))
      return; // done! (we hope ;-) )

    // loop over all the container's elements
    const koElements = oCont.GetElements();
    while(koElements.hasMoreElements())
    {
      const koElement = koElements.getNext();
      if (    (koElement instanceof gkIRDFResource)
           && (this.moRDFContainerUtils.IsContainer(this.moDataSource, koElement))
         )
      { // recur subcontainers
        this.RemoveContainer(koElement);
      }
      // do something with this element before erasing it
      if (afCallback)
        afCallback(koElement, aoUserData) ;
      // remove all outgoing arcs from Element
      this.RemoveArcsOut(koElement);
      oCont.RemoveElement(koElement, false);
    }
    // remove all arcs from this container
    // (this is a workaround to get rid of main level containers)
    this.RemoveArcsOut(koContainer);
  },


  // |
  // |  node functions
  // |

  // check if arsNode exists (in arsContainer if given)
  // (the name of arsNode is independent from arsContainer!)
  // will return the node resource or |null|
  ExistNode: function(arsNode, arsContainer)
  {
    var oNodeExists = null;
    if (arsNode)
    {
      const koNodeRes = this.GetNodeRes(arsNode);
      var oCont = null;
      if (arsContainer)
      { // we can use the container functions
        oCont = this.InitContainer(arsContainer);
      }

      if (oCont)
      { // use container function
        if (oCont.IndexOf(koNodeRes) != -1)
          oNodeExists = koNodeRes;
      }
      else
      { // we assume that a node does only exist if there are arcs in or out
        const koExist = {bExist:  false,
                         fExist:  function(ignored1, ignored2, ignored3, aoUserData)
                                  {
                                    if (aoUserData)
                                      aoUserData.bExist = true;
                                    return false;   // stop at first arc found
                                  }
                        };
        this.GetAllArcsOut (koNodeRes, koExist.fExist, koExist);
        if (!koExist.bExist)
          this.GetAllArcsIn(koNodeRes, koExist.fExist, koExist);
        if (koExist.bExist)
          oNodeExists = koNodeRes;
      }
    } // have node argument
    return oNodeExists;
  },


  // add node arsNode to arsContainer and fill with all elements
  // of the *associative* Array aroElementArray
  // (the name of arsNode is independent from arsContainer!)
  AddNode: function(arsNode, arsContainer, aroElementArray)
  {
    var oCont = this.InitContainer(arsContainer);
    if (oCont)
    {
      const koItem = this.GetNodeRes(arsNode);
      if (oCont.IndexOf(koItem) == -1)
        oCont.AppendElement(koItem);   // create new node
      for (var sElem in aroElementArray)
      {
        this.ChangeItemValue(aroElementArray[sElem],
                             sElem,
                             arsNode);
      }
    }
  },


  // remove the node arsNode by snipping all its arcs
  // aoNode should be urn array or nsIRDFResource
  RemoveNode: function(aoNode)
  {
    if (aoNode)
    {
      const koItem = this.GetNodeRes(aoNode);
      this.RemoveArcsOut(koItem);
      this.RemoveArcsIn (koItem);
    }
  },


  // get a list with all the container's node names
  // if possible, the node names will be relative to arsContainer
  // if given, the list is restricted to nodes with asItem set to asValue
  // if abAssoc is |true|, the returned array is associative, with all indices prepended
  //  by asPrefix and all values set to aoTag
  // if abAssoc is |false|, an array of strings is returned
  // if arNodeList is given, it will be expanded
  GetNodeList: function(arsContainer, abAssoc, asItem, asValue, arNodeList, asPrefix, aoTag)
  {
    if (!arNodeList)
      arNodeList = [];
    var asValue = String(asValue);
    if (!asPrefix)
      asPrefix = "";

    var oCont = this.InitContainer(arsContainer);
    if (oCont)
    {
      var nsISimpleEnumerator = oCont.GetElements();
      while (nsISimpleEnumerator.hasMoreElements())
      {
        var oNode = nsISimpleEnumerator.getNext();
        var bGetNode = !(asItem && asValue && asItem.length && asValue.length);
        if (!bGetNode)
        { // check the item value
          try
          {
            var oTarget = this.moDataSource.GetTarget(this.GetNodeRes(oNode),
                                                     this.GetItemRes(asItem),
                                                     true);
            if ((oTarget instanceof gkIRDFResource) || (oTarget instanceof gkIRDFLiteral))
              bGetNode = (String(target.Value) == String(asValue));
          }
          catch(ignore){}
        }

        if (bGetNode && ((oNode instanceof gkIRDFResource) || (oNode instanceof gkIRDFLiteral)))
        { // remove the container URN prefix, if possible
          const kxURNPrefix = new RegExp("^" + this.CreateURN(arsContainer) + ":"); // eg. ^urn:mnenhy
          var sNodeName = oNode.Value.replace(kxURNPrefix, "");
          if (abAssoc)
            arNodeList[asPrefix + sNodeName] = aoTag;
          else
            arNodeList.push(sNodeName);
        }
      } // while
    } // if container init ok
    return arNodeList;
  },


  // get nsISimpleEnumerator with all elements that have asItem point to arsNode
  GetSourcesForNode: function(asItem, arsNode)
  {
    return this.moDataSource.GetSources(this.GetItemRes(asItem),
                                       this.GetNodeRes(arsNode),
                                       true);
  },


  // |
  // |  item functions
  // |

  // get the value of the asItem of arsNode
  GetItemValue: function(asItem, arsNode)
  {
    var oVal = null;
    try
    {
      var oTarget = this.moDataSource.GetTarget(this.GetNodeRes(arsNode),
                                               this.GetItemRes(asItem),
                                               true);
      if ((oTarget instanceof gkIRDFResource) || (oTarget instanceof gkIRDFLiteral))
        oVal = oTarget.Value;
    }
    catch(ignore){ /* return only default null value */ }
    return oVal;
  },


  // change the value of asItem of arsNode to aoNewValue
  // if aoNewValue is empty, it will be deleted from the rdf
  ChangeItemValue: function(aoNewValue, asItem, arsNode)
  {
    var oOldValue = this.GetItemValue(asItem, arsNode);
    if (oOldValue)
      oOldValue  = this.GetLit(oOldValue);
    if (aoNewValue)
      aoNewValue = this.GetLit(aoNewValue);

    if (oOldValue || aoNewValue)
    {
      const koNode = this.GetNodeRes(arsNode);
      const koItem = this.GetItemRes(asItem);


      if (oOldValue && !aoNewValue)
      {
        this.moDataSource.Unassert(koNode, koItem, oOldValue);
      }
      else if (!oOldValue && aoNewValue)
      {
        this.moDataSource.Assert(koNode, koItem, aoNewValue, true);
      }
      else
      {
        this.moDataSource.Change(koNode, koItem, oOldValue, aoNewValue);
      }
    }
  },


  //
  // some convenience functions
  //

  GetRes: function(asResName)
  {
    return this.moRDFService.GetResource(asResName);
  },
  GetLit: function(asLiteral)
  {
    return this.moRDFService.GetLiteral(asLiteral);
  },

  // convert to node resource
  // aoNode should be urn array or nsIRDFResource
  GetNodeRes: function(aoNode)
  {
    if (aoNode instanceof Components.interfaces.nsIRDFResource)
      return aoNode;
    else
      return this.GetRes(this.CreateURN(aoNode));
  },

  // convert to item resource by adding $datanamespace to $item
  GetItemRes: function(asItemName)
  {
    return this.GetRes(this.msDataNameSpace + asItemName);
  },


  // check existence of container via nsIRDFContainer.Init
  // return the nsIRDFContainer object or null
  // aoContainer should be urn array or nsIRDFResource
  InitContainer: function(aoContainer)
  {
    const koContRes = this.GetNodeRes(aoContainer);
    try
    {
      return new this.mCRDFContainer(this.moDataSource, koContRes);
    }
    catch(error)
    { // we usually get here if sContainer does not exist
      this.debugdump(",-Fatal Error in MnenhyRDF::InitContainer(["+koContRes.Value+"])\n| "+error+"\n`-\n");
      return null;
    }
  },


  // create URI string of RDF datasource
  //  IN:     nsIURI or string    aoSource    the rdf file URI to be opened
  //  IN:     nsIURI or string    aoDefault   default file that will be copied if no aoSource exists
  //  RETURN: string                          the rdf file URI string or |null|
  // if no file exists and no default file could be copied, an empty file will be created
  GetDataSourceURI: function(aoSource, aoDefault)
  {
    // get a real file location: no paths into a jar allowed
    const ksURI    = (aoSource instanceof Components.interfaces.nsIURI) ? aoSource.spec : aoSource;
    const krSource = [];
    aoSource = goMnenhy.ConvertURI2FileURI(aoSource, krSource);
    if (krSource.length != 1)
      return ksURI; // URI spec that effectively describes a JAR URI

    // check existence
    if (!aoSource.file.exists())
    { // not found; maybe we can copy it from somewhere?
      const krDefault = [];
      aoDefault = goMnenhy.ConvertURI2FileURI(aoDefault, krDefault);

      var bCreate = !(aoDefault && aoDefault.file.exists());
      if (!bCreate)
      { // have default file
        if (krDefault.length > 1)
        { // need to extract from jar
          bCreate = !goMnenhy.ExtractZip(aoDefault.file, aoSource.file.parent, krDefault[1], true);
        }
        else
        { // copy it
          aoDefault.file.copyTo(aoSource.file.parent, aoSource.file.leafName);
          bCreate = !aoSource.file.exists();
        }
      } // have default file

      if (bCreate)
      { // create new aoSource
        const klPermissions = 0644;  // only owner may change, others only read
        try
        {
          aoSource.file.create(aoSource.file.NORMAL_FILE_TYPE, klPermissions)
        }
        catch(ignored)
        {
          return null;
        }
      } // create new aoSource
      this.mbDataSourceCreated = true;
    }
    return aoSource.spec;
  },


  // |
  // |  debug
  // |
  EnableDebugDump: function()
  {
    this.mbDebugDump = true;
  },
  DisableDebugDump: function()
  {
    this.mbDebugDump = false;
  },
  debugdump: function(asSomethingToDump)
  {
    if (this.mbDebugDump)
      dump("MnenhyRDF: "+asSomethingToDump);
  }

};

// default notifier function
// gets called if loading datasource is finished
function DefaultNotifier(aoMnenhyRDF, aoUserDataIgnored)
{ // do nothing but dump
  if (aoMnenhyRDF && aoMnenhyRDF.mbDebugDump)
  {
    const bLoaded = ((aoMnenhyRDF.moDataSource instanceof gkIRDFRemoteDataSource) && (aoMnenhyRDF.moDataSource.loaded));
    aoMnenhyRDF.debugdump("DefaultNotifier: loaded="+bLoaded+", "+aoMnenhyRDF.msDataSourceURI+"\n");
  }
}



// |
// |  Init called by the global Mnenhy object when loading this script file
// |
function Init(aoMnenhy)
{ // keep pointer to have the Mnenhy object prototype available
  this.goMnenhy = aoMnenhy;
}

